SystemOrganization addCategory: #'Croquet-Import'!

TestCase subclass: #TLoad3DSMaxTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Import'!

!TLoad3DSMaxTest methodsFor: 'Testing' stamp: 'ar 4/4/2006 21:13'!
testNextToken
	| strm tload ans |
	strm := '*3DSMAX_ASCIIEXPORT	200' readStream.
	tload := TLoad3DSMax basicNew reset.
	ans := tload nextOn: strm.
	self assert: ans first = #T3dsmaxAsciiexport.
	self assert: ans second = '200'.
	self assert: (ans := tload nextOn: strm) first = #EOF.
! !

!TLoad3DSMaxTest methodsFor: 'Testing' stamp: 'ar 4/4/2006 21:13'!
testNextTokenWithParenthesis
	| strm tload ans |
	strm := '	*MATERIAL 0 {
		*MATERIAL_NAME "plateau01"
		*LONG 1 2 3 *MULTI 4	5	6
	}' readStream.
	tload := TLoad3DSMax basicNew.
	ans := tload reset nextOn: strm.
	self assert: ans first = #TMaterial.
	self assert: ans second = '0'.

	ans := tload nextOn: strm.
	self assert: ans first = #BEGIN.

	ans := tload nextOn: strm.
	self assert: ans first = #TMaterialName.
	self assert: ans second = '"plateau01"'.

	ans := tload nextOn: strm.
	self assert: ans first = #TLong.
	self assert: ans second = '1 2 3'.

	ans := tload nextOn: strm.
	self assert: ans first = #TMulti.
	self assert: ans second = '4	5	6'.

	ans := tload nextOn: strm.
	self assert: ans first = #END.

	self assert: (ans := tload nextOn: strm) first = #EOF.
! !

!TLoad3DSMaxTest methodsFor: 'Testing' stamp: 'ar 4/4/2006 21:14'!
testParse
	| source tree |
	source := '*3DSMAX_ASCIIEXPORT	200'.
	tree := TLoad3DSMax basicNew parse: source readStream.
	self assert: tree first = #T3dsmaxAsciiexport.
	self assert: tree second = '200'.! !

!TLoad3DSMaxTest methodsFor: 'Testing' stamp: 'tak 6/23/2004 10:58'!
testParseNesting
	| source tree |
	source := '
*FIRST 1
*SECOND  2 {
	*THIRD 3 {
		*FOURTH 4
	}
}
*FIFTH 5
'.
	tree := TLoad3DSMax basicNew parse: source readStream.
	self assert: tree first = #TFirst.
	self assert: tree fourth last first = #TFourth.
! !

!TLoad3DSMaxTest methodsFor: 'Testing' stamp: 'tak 6/23/2004 10:58'!
testParseNesting2
	| source tree |
	source := '
*FIRST 1
*SECOND  2 {
	*THIRD 3 {
	}
}
*FIFTH 5
'.
	tree := TLoad3DSMax basicNew parse: source readStream.
	self assert: tree first = #TFirst.
	self assert: tree last = '5'.
! !

!LinkedList methodsFor: '*Croquet-Import' stamp: 'sn 2/26/2018 03:18'!
addLastLink: aLink
	"Add aLink to the end of the receiver's list. Answer aLink."
	self isEmpty
		ifTrue: [firstLink := aLink]
		ifFalse: [lastLink nextLink: aLink].
	lastLink := aLink.
	^aLink! !

Link subclass: #T3DSFaceNode
	instanceVariableNames: 'face'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Import'!

!T3DSFaceNode class methodsFor: 'as yet unclassified' stamp: 'jhk 5/1/2005 20:04'!
face: theFace
	"Create a new node with the face in it"

	|  |
	(self class new) face: theFace.! !

!T3DSFaceNode methodsFor: 'as yet unclassified' stamp: 'jhk 5/1/2005 20:02'!
face
	"Return the value for the face"

	|  |
	^face.! !

!T3DSFaceNode methodsFor: 'as yet unclassified' stamp: 'jhk 5/1/2005 20:02'!
face: value
	"Set the value for the face"

	|  |
	face := value.! !

Link subclass: #T3DSParseNode
	instanceVariableNames: 'name value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Import'!

!T3DSParseNode class methodsFor: 'as yet unclassified' stamp: 'jhk 4/27/2005 14:06'!
name: nme value: val
	"Create a LinkList node with name value pairs"

	| node |
	node := self new.
	node name: nme.
	node value: val.
	^node.! !

!T3DSParseNode methodsFor: 'as yet unclassified' stamp: 'sn 2/26/2018 03:33'!
asLink

	^ self! !

!T3DSParseNode methodsFor: 'as yet unclassified' stamp: 'jhk 4/27/2005 13:59'!
name
	"Return the name"

	|  |
	^name.! !

!T3DSParseNode methodsFor: 'as yet unclassified' stamp: 'jhk 4/27/2005 13:58'!
name: nme
	"Set the name"

	|  |
	name := nme.! !

!T3DSParseNode methodsFor: 'as yet unclassified' stamp: 'jhk 4/27/2005 13:59'!
value
	"Return the value"

	|  |
	^value.! !

!T3DSParseNode methodsFor: 'as yet unclassified' stamp: 'jhk 4/27/2005 13:58'!
value: val
	"Set the value"

	|  |
	value := val.! !

Object subclass: #TLoadOBJ
	instanceVariableNames: 'scale vertices normals textureUV faces materials frame'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Import'!

!TLoadOBJ class methodsFor: 'class initialization' stamp: 'pbm 11/15/2006 15:45'!
initializeWithFileName: fileName
	^ self new initializeWithFileName: fileName! !

!TLoadOBJ class methodsFor: 'class initialization' stamp: 'pbm 11/15/2006 15:45'!
initializeWithFileName: fileName scale: scl
	^ self new initializeWithFileName: fileName scale: scl! !

!TLoadOBJ methodsFor: 'meshing' stamp: 'pbm 2/15/2007 16:52'!
buildMesh
	| faceGroups materialList vertexBuffer normalBuffer texCoordBuffer index selectedFaces sz count indexBuffer |
	sz := faces size * 3.
	vertexBuffer := Vector3Array ofSize: sz.
	normalBuffer := Vector3Array ofSize: sz.
	texCoordBuffer := Vector2Array ofSize: sz.
	faceGroups := OrderedCollection new.
	materialList := OrderedCollection new.

	count := 0.
	materials keysAndValuesDo: [:matName :mat |
		materialList add: mat.
		faceGroups add: materialList size.

		selectedFaces := faces select: [:face | face materialName = matName].
		indexBuffer := IntegerArray ofSize: selectedFaces size * 3.

		selectedFaces withIndexDo: [:face :i |
			1 to: 3 do: [:j |
				index := ((i-1) * 3 + j).
				vertexBuffer at: count + index put: (vertices at: (face vertexIndices at: j)) * scale.
				normalBuffer at: count + index put: (normals at: (face normalIndices at: j)).

				face textureUVIndices isEmpty ifFalse: [
					texCoordBuffer at: count + index put: (textureUV at: (face textureUVIndices at: j)).
				].
				indexBuffer at: index put: count + index - 1.
			].
		].
		count := count + indexBuffer size.
		faceGroups add: indexBuffer.
	].

	^ TMesh 
		triMeshWithVertices: vertexBuffer
		alias: nil
		norms: normalBuffer
		textureUV: texCoordBuffer
		faceGroups: faceGroups
		material: (TMaterial new subMaterialList: materialList).! !

!TLoadOBJ methodsFor: 'accessing' stamp: 'pbm 10/21/2006 20:20'!
faces
	^ faces! !

!TLoadOBJ methodsFor: 'accessing' stamp: 'pbm 11/20/2006 16:33'!
frame
	^ frame ifNil: [ frame := self buildMesh ]! !

!TLoadOBJ methodsFor: 'initialize' stamp: 'pbm 3/14/2007 11:20'!
initialize
	super initialize.
	scale := 1.0.
	vertices := OrderedCollection new.
	normals := OrderedCollection new.
	textureUV := OrderedCollection new.
	faces := OrderedCollection new.
	materials := Dictionary new.! !

!TLoadOBJ methodsFor: 'initialize' stamp: 'pbm 11/15/2006 14:43'!
initializeWithFileName: fileName
	self initializeWithFileName: fileName scale: 1.0.! !

!TLoadOBJ methodsFor: 'initialize' stamp: 'pbm 11/15/2006 17:10'!
initializeWithFileName: fileName scale: scl
	| fileStream |

	" Load the .obj file. "
	fileStream := (CrLfFileStream readOnlyFileNamed: fileName) ascii.
	scale := scl.

	[self parseStream: fileStream] ensure: [fileStream close].! !

!TLoadOBJ methodsFor: 'parsing' stamp: 'pbm 10/25/2006 19:32'!
loadTexture: fileName

	^ TTexture new initializeWithFileName: fileName mipmap: true shrinkFit: false! !

!TLoadOBJ methodsFor: 'parsing' stamp: 'pbm 11/16/2006 11:31'!
makeColor: args

	^ VectorColor
			r: args second asNumber
			g: args third asNumber
			b: args fourth asNumber
			a: 1.0! !

!TLoadOBJ methodsFor: 'parsing' stamp: 'pbm 9/4/2007 10:51'!
makeFace: args

	| face indices |
	face := TOBJFace new.

	2 to: args size do: [:i |
		indices := (args at: i) findTokens: $/.
		indices size = 1 ifTrue: [ 
			face addVertex: (indices at: 1) texture: nil normal: nil.
		].
		indices size = 2 ifTrue: [
			((args at: i) includesSubString: '//') ifTrue: [
				face addVertex: (indices at: 1) texture: nil normal: (indices at: 2).
			] ifFalse: [
				face addVertex: (indices at: 1) texture: (indices at: 2) normal: nil.
			].
		].
		indices size = 3 ifTrue: [
			face addVertex: (indices at: 1) texture: (indices at: 2) normal: (indices at: 3).
		].
	].

	^ face! !

!TLoadOBJ methodsFor: 'parsing' stamp: 'pbm 11/16/2006 17:29'!
makeNormal: args
	| normal |
	normal := Vector3
				x: (args at: 2) asNumber
				y: (args at: 3) asNumber
				z: (args at: 4) asNumber.

	normals add: normal.! !

!TLoadOBJ methodsFor: 'parsing' stamp: 'pbm 11/14/2006 15:55'!
makeTextureUV: args
	textureUV add: (Vector2
						u: (args at: 2) asNumber
						v: 1.0 + (0.0 - (args at: 3) asNumber)).! !

!TLoadOBJ methodsFor: 'parsing' stamp: 'pbm 10/21/2006 20:17'!
makeVector3: fileStream

	^ Vector3
		x: (fileStream upTo: Character space) asNumber
		y: (fileStream upTo: Character space) asNumber
		z: (fileStream upTo: Character cr) asNumber! !

!TLoadOBJ methodsFor: 'parsing' stamp: 'pbm 11/14/2006 15:51'!
makeVertex: args
	| vertex |
	vertex := Vector3
				x: (args at: 2) asNumber
				y: (args at: 3) asNumber
				z: (args at: 4) asNumber.
	vertices add: vertex.! !

!TLoadOBJ methodsFor: 'accessing' stamp: 'pbm 10/25/2006 14:33'!
materials
	^ materials! !

!TLoadOBJ methodsFor: 'accessing' stamp: 'pbm 11/14/2006 15:44'!
normals
	^ normals! !

!TLoadOBJ methodsFor: 'parsing' stamp: 'pbm 11/22/2006 14:41'!
parseMtlFile: fileName

	| fileStream tokens currentMaterial localDir file |
	fileStream := (CrLfFileStream readOnlyFileNamed: fileName) ascii.
	localDir := fileStream directory pathName.

	[ fileStream atEnd ] whileFalse: [
		tokens := fileStream nextLine findTokens: Character separators.
		tokens isEmpty ifFalse: [
			tokens first asSymbol caseOf: {
				[ #newmtl ] -> [ "New Material"
					tokens second ifNotNilDo: [:name |
						currentMaterial := materials at: name ifAbsentPut: TMaterial new.
					].
				].

				[ #Ns ] -> [ "Shininess"
					currentMaterial shininess: tokens second asNumber asFloat.
				].

				[ #Ka ] -> [ "Ambient Color"
					currentMaterial ambientColor: (self makeColor: tokens).
				].

				[ #Kd ] -> [ "Diffuse Color"
					currentMaterial diffuseColor: (self makeColor: tokens).
				].

				[ #Ks ] -> [ "Specular Color"
					currentMaterial specularColor: (self makeColor: tokens).
				].

				[ #'map_Kd' ] -> [ "Texture Map"
					file := FileDirectory pathFrom: {localDir. tokens last}.
					currentMaterial texture: (self loadTexture: file).
				].
					
			} otherwise: [].
		].
	].! !

!TLoadOBJ methodsFor: 'parsing' stamp: 'pbm 3/14/2007 11:20'!
parseStream: fileStream

	| face localDir txt useMaterial tokens currentMaterial |
	localDir := fileStream directory pathName.

	" Loop through each line of the file. "
	[ fileStream atEnd ] whileFalse: [
		tokens := fileStream nextLine findTokens: Character separators.
		tokens isEmpty ifFalse: [
			tokens first asSymbol caseOf: {
				[ #v ] -> [ "Vertex"
					self makeVertex: tokens.
				].

				[ #vn ] -> [ "Vertex Normals"
					self makeNormal: tokens.
				].

				[ #vt ] -> [ "Texture Coordinates"
					self makeTextureUV: tokens.
				].

				[ #g ] -> [ "Group"
					"Ignoring for now"
				].

				[ #mtllib ] -> [ "Material Library"
					self parseMtlFile: (FileDirectory pathFrom: {localDir. tokens second}).
				].

				[ #usemtl ] -> [ "Use Material"
					useMaterial := tokens second.
				].

				[ #usemap ] -> [ "Use Texture Map"
					txt := self loadTexture: (FileDirectory pathFrom: {localDir. tokens second}).
					currentMaterial := materials at: useMaterial ifAbsentPut: [ TMaterial new ].
					currentMaterial ifNotNilDo: [:mat | mat texture: txt].
				].

				[ #f ] -> [ "Face"
					face := self makeFace: tokens.
					useMaterial ifNotNilDo: [:m | face materialName: m].
					faces add: face.
				].
			} otherwise: [].
		].
	].

	self triangulateFaces.

	" Print some info. "
	true ifTrue: [
		Transcript show: 'Vertices read: ', vertices size; cr.
		Transcript show: 'Normals read: ', normals size; cr.
		Transcript show: 'TextureUV read: ', textureUV size; cr.
		Transcript show: 'Faces read: ', faces size; cr.
	].! !

!TLoadOBJ methodsFor: 'accessing' stamp: 'pbm 11/21/2006 15:01'!
textureUV
	^ textureUV! !

!TLoadOBJ methodsFor: 'parsing' stamp: 'pbm 11/21/2006 17:29'!
triangulateFaces

	| newFaces |
	newFaces := OrderedCollection new.

	faces do: [:face | newFaces addAll: face triangulate ].
	faces := newFaces.! !

!TLoadOBJ methodsFor: 'accessing' stamp: 'pbm 10/20/2006 15:00'!
vertices
	^ vertices! !

Object subclass: #TOBJFace
	instanceVariableNames: 'vertexIndices normalIndices textureUVIndices materialName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Import'!

!TOBJFace methodsFor: 'accessing' stamp: 'pbm 11/20/2006 16:51'!
addVertex: vtx texture: txt normal: nrml
	vertexIndices add: vtx asNumber.
	txt ifNotNil: [ textureUVIndices add: txt asNumber ].
	nrml ifNotNil: [ normalIndices add: nrml asNumber ].! !

!TOBJFace methodsFor: 'initialize' stamp: 'pbm 11/20/2006 16:48'!
initialize
	vertexIndices := OrderedCollection new.
	normalIndices := OrderedCollection new.
	textureUVIndices := OrderedCollection new.! !

!TOBJFace methodsFor: 'accessing' stamp: 'pbm 11/20/2006 16:53'!
materialName
	^ materialName! !

!TOBJFace methodsFor: 'accessing' stamp: 'pbm 11/20/2006 16:53'!
materialName: matName
	materialName := matName.! !

!TOBJFace methodsFor: 'accessing' stamp: 'pbm 11/20/2006 16:52'!
normalIndices
	^ normalIndices! !

!TOBJFace methodsFor: 'accessing' stamp: 'pbm 11/20/2006 16:54'!
size
	^ vertexIndices size! !

!TOBJFace methodsFor: 'accessing' stamp: 'pbm 11/20/2006 16:52'!
textureUVIndices
	^ textureUVIndices! !

!TOBJFace methodsFor: 'accessing' stamp: 'pbm 11/21/2006 14:34'!
triangulate
	| newFaces newFace |

	"Returns a new collection of triangular faces if this face has > 3 vertices."
	newFaces := OrderedCollection new.
	vertexIndices size = 3 
		ifTrue: [ newFaces add: self ]
		ifFalse: [
	 		3 to: vertexIndices size do: [:i |
				newFace := self class new.
				newFace materialName: materialName.

				{1. i-1. i} do: [:j |
					newFace vertexIndices add: (vertexIndices at: j).
					textureUVIndices size > 0 ifTrue: [
						newFace textureUVIndices add: (textureUVIndices at: j).
					].
					normalIndices size > 0 ifTrue: [
						newFace normalIndices add: (normalIndices at: j).
					].
				].
			
				newFaces add: newFace.
			].
		].

	^ newFaces
! !

!TOBJFace methodsFor: 'accessing' stamp: 'pbm 11/20/2006 16:52'!
vertexIndices
	^ vertexIndices! !

Object subclass: #TLoad3DSMax
	instanceVariableNames: 'vertices tvertices faces tfaces matIndex stringLoc materialList materialRef textureList scale shadeAngle shadeAngleCos frame textureMode transform inverseTransform aheadChar filePath meshFaceNormals meshFaceVertexNormals objectName localTransform'
	classVariableNames: 'DelimSet1 DelimSet2 DelimSet3 NumberSet WhiteSpaceSet'
	poolDictionaries: 'OpenGLConstants'
	category: 'Croquet-Import'!

!TLoad3DSMax commentStamp: 'das 10/10/2004 12:21' prior: 0!
TLoad3DSMax Notes:

There are two parts to the class - the parse routines and everything else. The actual #parse: method creates a block hierarchy - essentially a kind of parse tree, from the inputs. 3DS is somewhat regular in its construction, but there are a few things to be careful about. The tree that gets constructed is made up of a field name and a field. The field names are tokenized 3DS field names and the field is the child tree or the actual text field data - not parsed into actual numbers or anything yet. One improvement on performance might be to interleave this step with the next.

Once the tree exists, a second pass is made that converts the text fields and constructs the actual frame hierarchy. The field name of each node is matched and the appropriate #make****: routine is called on its contents. 

Once a raw mesh is set up inside the #makeGeometry: method, the #reconstruct: method is called. This does a lot of massaging and optimization of the raw vertices - including aliasing, etc. 

Anyway, start at:
	#initialize: fileName: scale: shadeAngle: textureMode: 

Based upon the field names in the tree constructed in #parse: you just make the call to the next #makeXXX:.

- Though we have all of the transforms for the heirarchy, the actual locations of the meshes is already transformed. If I can think of a good reason to un-transform the mesh elements and then add the transform to the actual frame, I will do it. For now, treat it as a solid body.

- The texture rotations are face centered. This requires an offset of 0.5, the rotation, and then putting it back. Not sure if I really want to pre-transform the texture uv coordinates. Also, not sure if this is already done, as the meshes are.

- This class will act as a template for importers, though they all seem so different that this may be easier said than done.

- This object is actually never rendered, and really doesn't need to be a subclass of TFrame.

DAS!

!TLoad3DSMax class methodsFor: 'class initialization' stamp: 'ar 6/16/2005 22:15'!
findFileName: fName

	| path |
	(FileStream isAFileNamed: fName) ifTrue:[^ fName].
	path := FileDirectory pathFromURI: fName.
	(FileStream isAFileNamed: path) ifTrue:[^path].
	path := FileDirectory pathFromURI: 'Content/', fName.
	(FileStream isAFileNamed: path) ifTrue:[^path].
	^ nil.! !

!TLoad3DSMax class methodsFor: 'class initialization' stamp: 'bgf 6/25/2006 11:31'!
initialize
	"TLoad3DSMax initialize"
	DelimSet1 := CharacterSet newFrom: '*{}'.
	DelimSet2 := CharacterSet newFrom: ' {}	',Character cr asString,Character lf asString.
	DelimSet3 := CharacterSet newFrom: '*}'.
	NumberSet := CharacterSet newFrom: '-0.123456789'.  "normals start with - or 0"
	WhiteSpaceSet :=  CharacterSet newFrom: (Array with: Character space with: Character tab).
! !

!TLoad3DSMax methodsFor: 'convert' stamp: ''!
aliasIndexToArray: ai

	| alias |

	alias := IntegerArray ofSize: ai size.
	1 to: ai size do:[ :i |
		(ai at: i) ifNil:[alias at: i put: 0.] 
				ifNotNil:[alias at: i put: (ai at: i).].
		].

	^ alias.! !

!TLoad3DSMax methodsFor: 'as yet unclassified' stamp: 'bgf 6/28/2006 17:26'!
applyOpacity: omap to: dmap

	"Oform is a MAP_OPAQUE from the model; dform is the MAP_DIFFUSE.
     Composite them to produce a 32-bit image that is transparent 
	for all pixels where oform is black,
	and a copy of dform's pixels for all pixels where oform is white."

	| result blit |

	"All that is needed is to apply the R (or G or B) value
     of each pixel from oform into dform as its alpha component."
 
	result := dmap asFormOfDepth:32.

	"Clear the alpha component of the result."
	blit := BitBlt toForm: result.
	blit combinationRule: 1;  "AND"
		halftoneForm: (Bitmap with: 16r00FFFFFF);
		copyBits.

	"Copy in the alpha from the blue of the map."
	blit := BitBlt toForm: result.
	blit sourceForm: (omap asFormOfDepth:32);
		combinationRule: 7; "OR"
		colorMap: (ColorMap
					shifts: #(0 0 0 24)
					masks: #(0 0 0 16rFF));
		copyBits.

	^ result.
! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'das 1/12/2006 10:20'!
basicCalcFaceNormals

"Calculate the normal for each primitive (simple vertex index) face."
	| v1 v2 v count faceNormals |

	faceNormals := Vector3Array  new: (faces size//3).
	count := 1.
	1 to: faceNormals size do:[   :i |
		v1 := (vertices at:(faces at: count)+1) - (vertices at: (faces at: count+1)+1).
		v2 := (vertices at:(faces at: count+1)+1) - (vertices at: (faces at: count+2)+1).
		v := v1 cross: v2.
		v length = 0.0 ifFalse:[ v:= v normalized.].
		faceNormals at: i put: v.
		count := count+3.].

	^ faceNormals.! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'das 1/12/2006 12:47'!
basicCalcVertexNormals: faceNormals

"Calculate the normal for each face vertex."
	| vertexNormals vertexCount iCount faceIndex1 faceIndex2 fv1 fv2 edges index1 |
"Initialize the vertexNormals with the owning face normal."
	vertexNormals := Vector3Array  new: (faces size).
	vertexCount := IntegerArray new: (faces size).
	iCount := 1.
	1 to: faces size by: 3 do:[ :i |
		vertexNormals at: i put: (faceNormals at: iCount).
		vertexNormals at: i+1 put: (faceNormals at: iCount).
		vertexNormals at: i+2 put: (faceNormals at: iCount).
		iCount := iCount+1.].
"vertexCount is set to zero at the moment. It should be 1, but we know that it is always 1 greater than whatever our end result will be below."
	"self halt."

	1 to: faces size do:[:i |
		i+1 to: faces size do:[:j |
			(faces at: i) = (faces at: j) ifTrue:[
				faceIndex1 := 1 + ((i-1)//3).
				faceIndex2 := 1 + ((j-1)//3).
				(self testNorm: (faceNormals at: faceIndex1) norm: (faceNormals at: faceIndex2))
					ifTrue:[
						vertexNormals at: i put: ((vertexNormals at: i) + (faceNormals at: faceIndex2)).
						vertexCount at: i put: 1+ (vertexCount at: i).
						vertexNormals at: j put: ((vertexNormals at: j) + (faceNormals at: faceIndex1)).
						vertexCount at: j put: 1+ ( vertexCount at: j).
						].
				].
			].
		].
false ifTrue:[
"Build an edge dictionary"
	edges := Dictionary new.
	1 to: faces size by: 3 do:[ :i |
		edges at: (faces at:i)@(faces at:i+1) put: i@(i+1).
		edges at: (faces at: i+1)@(faces at: i+2) put: (i+1)@(i+2).
		edges at: (faces at: i+2)@(faces at: i) put: (i+2)@i].

"Now match edges"
	faceIndex2 := 1.
	1 to: faces size by: 3 do:[ :i |
		index1 := edges at: ((faces at:i+1)@(faces at:i)) ifAbsent:[ nil ].
		index1 ifNotNil:[ fv1 := i+1. fv2 := i.] ifNil:[
			index1 := edges at: ((faces at: i+2)@(faces at: i+1)) ifAbsent:[ nil. ].
			index1 ifNotNil:[ fv1 := i+2. fv2 := i+1.] ifNil:[
				index1 := edges at: ((faces at: i)@ (faces at: i+2)) ifAbsent:[ nil. ].
				index1 ifNotNil:[fv1 := i. fv2 := i+2.].
			].].

		index1 ifNotNil:[
			faceIndex1 := 1+((index1 x-1)//3).
			(self testNorm: (faceNormals at: faceIndex1) norm:(faceNormals at: faceIndex2))
				ifTrue:[
					vertexNormals at: fv1 put: ((vertexNormals at: fv1) + (faceNormals at: faceIndex1)).
					vertexNormals at: fv2 put: ((vertexNormals at: fv2) + (faceNormals at: faceIndex1)).
					vertexCount at: fv1 put: 1+(vertexCount at: fv1).
					vertexCount at: fv2 put: 1+(vertexCount at: fv2).
					fv1 := index1 x.
					fv2 := index1 y.
					vertexNormals at: fv1 put: ((vertexNormals at: fv1) + (faceNormals at: faceIndex2)).
					vertexNormals at: fv2 put: ((vertexNormals at: fv2) + (faceNormals at: faceIndex2)).].
					vertexCount at: fv1 put: 1+(vertexCount at: fv1).
					vertexCount at: fv2 put: 1+(vertexCount at: fv2).
				].						
		faceIndex2 := faceIndex2 + 1.
		].
].
" Calculate the average normal and normalize."
	1 to: vertexNormals size do:[ :i |
		(vertexNormals at: i) length = 0.0 ifFalse:[
		vertexNormals at:i put: ((vertexNormals at: i)/(1.0 + (vertexCount at: i)))normalized.].].

	^ vertexNormals.! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'das 5/16/2006 14:51'!
calcFaceNormals

"Calculate the normal for each primitive (simple vertex index) face."
	| faceNormals count v1 v2 v midVertex |

	faceNormals := Vector3Array  new: (faces size//3).
	(self primCalcFaceNormals: vertices faces: faces faceNormals: faceNormals) 
		ifNotNil:[^faceNormals].

	count := 1.
	1 to: faceNormals size do:[ :i |
		
		midVertex := (vertices at: (faces at: count+1)+1).  "performance"
		v1 := (vertices at:(faces at: count)+1) - midVertex.
		v2 := midVertex - (vertices at: (faces at: count+2)+1).
		v := v1 cross: v2.
		v length = 0.0 ifFalse:[ v := v normalized.].
		faceNormals at: i put: v.
		count := count+3.
		].

	^ faceNormals.! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'das 1/12/2006 10:19'!
calcTextureVertices

" This is a somewhat naive fix to the texture transform issue. Be aware that I am not checking for shared texture coordinates. This means that if two textures share the same uv coordinate before the transform but not after, we have a problem. I don't know what 3DS Max does, so I will deal with it when I find the problem."
	| tv ml counter tv2 fvi avi ai found oc |

	ml := materialList at: 1+ materialRef.
	ml subMaterialList ifNotNil:[ ml := ml subMaterialList.] 
	ifNil:[
		oc := OrderedCollection new.
		oc add: ml.
		ml := oc.
		].
	tv := Vector2Array ofSize: tfaces size.
	1 to: tfaces size do:[ :i |
		counter := 1+(i-1//3).
		
		tv at: i put: (self textureTransform: (tvertices at: 1+(tfaces at: i)) 
			material: (ml at: 1+((matIndex at: counter)\\ml size))).
		].

	tv2 := OrderedCollection ofSize: tvertices size.
	ai := OrderedCollection ofSize: tvertices size.
	1 to: tfaces size do:[ :i | 
		found := false.
		fvi := 1+ (tfaces at: i).
		avi := fvi.
		[ found ] whileFalse:[
			"Anything already in this slot?"
			(tv2 at: avi) ifNil:[
				tv2 at: avi put: (tv at:i).
				tfaces at: i put: avi -1.
				found := true.].
			"Is it the same thing we already have?"
			found ifFalse:[
				(tv2 at: avi) = ( tv at: i) ifTrue:[
					tfaces at: i put: avi-1.
					found := true.].].
			"There is something already there, but it is the wrong thing, we need an alias"
			found ifFalse:[
				(ai at: avi) ifNil:[
					ai add: nil.
					tv2 add: (tv at: i).
					tfaces at: i put: (tv2 size -1). "index of new element - 1"
					ai at: avi put: tv2 size. "pointer to new element"
					found := true.] 
				"The alias was full, check the next alias."
				ifNotNil:[ avi := ai at: avi.].].].].

	" Convert array."
	tvertices := Vector2Array ofSize: tv2 size.
	1 to: tv2 size do:[ :i | (tv2 at: i) ifNotNil:[ tvertices at: i put: (tv2 at: i).].].
				! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'das 7/21/2006 21:18'!
calcVertexNormals: faceNormals

"Calculate the normal for each face vertex."
	| vertexNormals vertexCount vertexFaces fIndex vf vfn |

	"This is O(n-squared) in faces".
	faces size > 1000 ifTrue: [
		Transcript cr; show: 'Calculating vertex normals (', faces size, ' faces).'
	].

"Initialize the vertexNormals with the owning face normal."
	vertexNormals := Vector3Array  new: (vertices size).
	vertexCount := IntegerArray new: (vertices size).
	(self primCalcVertexNormals: vertexNormals vertexCount: vertexCount faces: faces faceNormals: faceNormals shadeAngleCos: shadeAngleCos)
		ifNotNil:[^vertexNormals].

	vertexFaces := Array ofSize: (vertices size).
	1 to: faces size do:[ :i |
		fIndex := (faces at:i)+1.
		(vertexFaces at: fIndex) ifNil:[vertexFaces at: fIndex put: OrderedCollection new.].
		(vertexFaces at: fIndex) add: 1 + ((i-1)//3).
	].

	1 to: vertexFaces size do:[ :i |
		vf := vertexFaces at:i.
		vf ifNotNil:[
			1 to: vf size do:[ :j |
				j+1 to: vf size do:[ :k |
					(self testNorm: (faceNormals at:(vf at:j)) norm: (faceNormals at: (vf at: k))) 
					ifTrue:[
						vertexNormals at: i put: (vertexNormals at: i)+(faceNormals at:(vf at:j)).
						vertexNormals at: i put: (vertexNormals at: i)+(faceNormals at:(vf at:j)).
						vertexCount at: i put: (vertexCount at: i)+2.
						].
					].
				].
			(vertexNormals at:i) length = 0 ifTrue:[
				1 to: vf size do:[ :m |
					vertexNormals at: i put: (vertexNormals at: i)+(faceNormals at:(vf at:m)).
					vertexCount at: i put: (vertexCount at: i)+1.
				]
			].
		].
	].
				

" Calculate the average normal and normalize."
	1 to: vertexNormals size do:[ :i |
		(vertexNormals at: i) length = 0.0 ifFalse:[
		vertexNormals at:i put: ((vertexNormals at: i)/(1.0 + (vertexCount at: i)))normalized.].].

	vfn := Vector3Array  new: (faces size).
	1 to: faces size do:[ :i |
		vfn at:i put: (vertexNormals at:(faces at:i)+1).
	].
	^ vfn.! !

!TLoad3DSMax methodsFor: 'parse' stamp: ''!
compare: v1 and: v2 within: epsilon

	| delta |
	delta := v1-v2.
	delta x abs > epsilon ifTrue:[^ false].
	delta y abs > epsilon ifTrue:[^ false].
	delta z abs > epsilon ifTrue:[^ false].
	^ true.! !

!TLoad3DSMax methodsFor: 'as yet unclassified' stamp: 'jhk 5/11/2005 08:09'!
compareB3DVector3: a to: b
	"compare two B3DVectors for similar values"

	|  |
	
	(((a x - b x) abs < 0.1) and: [((a y - b y) abs < 0.1) and:  [((a z - b z) abs < 0.1)]]) ifFalse: [
		Transcript cr; show: 'values not close '; cr.
		Transcript show: a asString; cr.
		Transcript show: b asString; cr.
		^false.
		].
	^true. ! !

!TLoad3DSMax methodsFor: 'as yet unclassified' stamp: 'jhk 5/11/2005 08:11'!
compareFileNormals: file computed: computed
	"compare two arrays for similar values"

	| errors |
	((file size) = (computed size)) ifFalse: [Transcript show: 'file size: ', file size asString, ' computed size: ', (computed size) asString; cr.
		^false.
		].
	errors := 0.
	1 to: (file size) do: [ :i | 
		(self compareB3DVector3: (file at: i) to: (computed at: i)) ifFalse: [
			Transcript show: 'at index: ', (i asString); cr. 
			errors := errors +1. 
			(errors > 5) ifTrue: [^false.].
			].
		].
	^true.! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'ar 4/4/2006 21:13'!
convertTName: aString 
	"(self new convertTName: 'SCENE_BACKGROUND_STATIC') = #TSceneBackgroundStatic"
	| strm start end str size name |
	name := aString asUppercase.
	name = 'MESH_VERTEX' ifTrue: [^#TMeshVertex].
	name = 'MESH_FACENORMAL' ifTrue: [^#TMeshFacenormal].
	name = 'MESH_VERTEXNORMAL' ifTrue: [^#TMeshVertexnormal].
	name = 'MESH_FACE' ifTrue: [^#TMeshFace].
	name = 'MESH_VERTCOL' ifTrue: [^#TMeshVertcol].
	name = 'MESH_CFACE' ifTrue: [^#TMeshCface].
	
	strm := WriteStream
				on: (String new: 40).
	strm nextPut: $T.
	"Because this is Tea and because some field start with  
	numbers."
	start := 1.
	size := aString size.
	[start <= size]
		whileTrue: [end := (aString
						indexOf: $_
						startingAt: start
						ifAbsent: [size + 1])
						- 1.
			str := aString copyFrom: start to: end.
			str translateToLowercase.
			str at: 1 put: str first asUppercase.
			strm nextPutAll: str.
			start := end + 2].
	^ strm contents asSymbol! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'das 4/4/2002 01:40'!
facesToFaceGroup: fcs materials: mi

	| groups count mIndex gIndex |
	
" This method creates an OrderedCollection of alternating material indices followed by the face group."

	groups := OrderedCollection new.

	count := 1.
	1 to: fcs size by: 3 do:[ :i |
		mIndex := 1+ (mi at: count).
		count := count+1.
		gIndex := 0.
		" Which material group do I belong to?"
		1 to: groups size by: 2 do:[ :j |
			(mIndex = (groups at: j)) ifTrue:[gIndex :=j].].
		"No such material group, we need to add a new one."
		gIndex = 0 ifTrue:[ 
			groups add: mIndex. 
			groups add: OrderedCollection new. 
			gIndex := groups size -1.].
		gIndex := gIndex+1.
		(groups at: gIndex) add: (fcs at:i).
		(groups at: gIndex) add: (fcs at: i+1).
		(groups at: gIndex) add: (fcs at: i+2).].
			
	1 to: groups size by: 2 do:[ :i |
		groups at: i+1 put:(groups at: i+1) asIntegerArray.].
	^ groups.! !

!TLoad3DSMax methodsFor: 'initialize' stamp: 'das 1/12/2006 10:45'!
findContentFileName: fName

	| path file |
	path := FileDirectory default fullNameFor: fName.
	(FileStream isAFileNamed: path) ifTrue:[^path].
	file := FileDirectory localNameFor: path.
	path := FileDirectory pathFromURI: 'Content/', file.
	(FileStream isAFileNamed: path) ifTrue:[^path].
	^ nil.! !

!TLoad3DSMax methodsFor: 'accessing' stamp: 'jcg 5/18/2005 16:30'!
frame

	frame class = TMesh ifTrue:[^ frame.].
	frame frameChildren ifNotNil:[
		frame frameChildren size = 1 ifTrue:[ ^ frame frameChildren at: 1. ].].
	^ frame.! !

!TLoad3DSMax methodsFor: 'initialize' stamp: 'DPR 8/21/2003 17:13'!
initializeWithFileName: fName 
	
	^self initializeWithFileName: fName scale: 1.0 shadeAngle: 90.1 textureMode: GLModulate
! !

!TLoad3DSMax methodsFor: 'initialize' stamp: 'DPR 8/21/2003 17:13'!
initializeWithFileName: fName scale: scl
	
	^self initializeWithFileName: fName scale: scl shadeAngle: 90.1 textureMode: GLModulate
! !

!TLoad3DSMax methodsFor: 'initialize' stamp: 'DPR 8/21/2003 17:13'!
initializeWithFileName: fName scale: scl shadeAngle: angle
	
	^self initializeWithFileName: fName scale: scl shadeAngle: angle textureMode: GLModulate.
! !

!TLoad3DSMax methodsFor: 'initialize' stamp: 'sn 2/26/2018 04:03'!
initializeWithFileName: fName scale: scl shadeAngle: angle textureMode: txtrMd
	
	| fieldName field parseTree mesh fileName |
	self class initialize.  "Make sure most recent delimiters are set"
	Smalltalk garbageCollectMost.
	scale := scl.
	textureMode := txtrMd.
	"get full path name of file, and path of directory containing it"
	fileName := self findContentFileName: fName.
	filePath := FileDirectory on: (FileDirectory dirPathFor: fileName).

Transcript show: fName; cr.
	parseTree := self parseFileNamed: fName.
	Smalltalk garbageCollectMost.
	self shadeAngle: angle. "Use degrees, but convert to radians"
	frame := TGroup new.
"Transcript show:'parsing';cr."
	parseTree linksDo:[ :node |
	
Transcript show:'.'.
		fieldName := node name.
		field := node value.
		fieldName = #T3dsmaxAsciiexport ifTrue:[ " unused "].
		fieldName = #TComment ifTrue:[" unused "].
		fieldName = #TScene ifTrue:[" unused "].
		fieldName = #TMaterialList ifTrue:[ self makeMaterials: field ].
		fieldName = #TGeomobject ifTrue:[ 
			mesh := (self makeGeometry: field).
			frame addChild: mesh.
			].
		fieldName = #TGroup ifTrue:[ 
			mesh := (self makeGroup: field). 
			frame addChild: mesh.
			].
		].
"Transcript show:'!!';cr. "
	frame frameChildren ifNotNilDo: 
		[:children | children size = 1 ifTrue: [frame := children at: 1]].
	frame prune.
	frame forceGlobalToLocal.
	Smalltalk garbageCollectMost.
Transcript show:'Loaded.';cr.! !

!TLoad3DSMax methodsFor: 'accessing' stamp: 'das 1/12/2006 11:01'!
localTransform

	^ localTransform.! !

!TLoad3DSMax methodsFor: 'accessing' stamp: 'das 1/12/2006 11:01'!
localTransform: lt

	localTransform := lt.! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 15:02'!
makeFaceNormals: tree
	"Extract the Mesh:=Face:=Normals from the tree"
	| meshFaceNormal fieldName tarray i n |

	meshFaceNormal := Vector3Array new: ((tree size)//4).
	i := 1.
	tree linksDo:[:node |
		fieldName := node name.

		fieldName = #TMeshFacenormal ifTrue:[
				tarray := self makeNumberArray: (node value).
				n := (Vector3 x:(tarray at: 2) y: (tarray at: 4) z:((tarray at: 3) negated)).
				n length = 0.0 ifFalse: [n := n normalized.].
				meshFaceNormal at: i put:  n.
				i := i + 1.
				].
		].
	^ meshFaceNormal.
! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 15:02'!
makeFaceVertexNormals: tree
	"Extract the face Mesh:=Vertex:=Normals from the tree"

	| meshFaceVertexNormal fieldName tarray i n |

	meshFaceVertexNormal := Vector3Array new: ((tree size)//4*3).
	i := 1.
	tree linksDo:[:node |
		fieldName := node name.
		fieldName = #TMeshVertexnormal ifTrue:[
				tarray := self makeNumberArray: (node value).
				n := (Vector3 x:(tarray at: 2) y: (tarray at: 4) z:((tarray at: 3) negated)).
				n length = 0.0 ifFalse: [n:= n normalized.].
				meshFaceVertexNormal at: i put: n.
				i := i + 1.
				].
		].
	^ meshFaceVertexNormal.
! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 03:52'!
makeGeometry: tree
	"handle the geometry fields"
	| tframe  fieldName start end children isMesh mesh |
	isMesh := false.
	children := OrderedCollection new.
	tree linksDo:[ :node |   "jhk altered to linked list for performance"
		fieldName := node name.
		fieldName = #THelperobject ifTrue:[
			tframe := TGroup new.
			tframe objectName: (self makeHelperObject: node value).
			tframe localTransform: transform.
			].
		fieldName = #TNodeName ifTrue:[ 
			start := (node value) findDelimiters: '"' startingAt: 1.
			end := (node value) findDelimiters: '"' startingAt: start+1.
			self objectName: ((node value) copyFrom: start+1 to: end-1).
			].
		fieldName = #TNodeTm ifTrue:[ self localTransform: (self makeTransform: node value).].
		fieldName = #TMesh ifTrue:[ 
			self makeMesh: node value.
			vertices size > 0 ifTrue:[isMesh := true.].
			].
		fieldName = #TMaterialRef ifTrue:[ materialRef := self makeNumber: node value.]. 
		].

	isMesh ifTrue:[ mesh := self reconstruct.] 
			ifFalse:[ mesh := TGroup new.
					mesh objectName: (self objectName).
					mesh localTransform: self localTransform.
                       ].
	children do:[ :c | mesh addChild: c].
	^ mesh.
		! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 03:53'!
makeGroup: tree

	| fieldName tframe |

	tframe := TGroup new.

	tree linksDo:[ :node |  "jhk altered to linked list"
		fieldName := node name.
		fieldName = #THelperobject ifTrue:[
			tframe objectName: (self makeHelperObject: node value).
			tframe localTransform: transform.
			].
		fieldName = #TGroup ifTrue:[
			tframe addChild: (self makeGroup: node value).
			].
		fieldName = #TGeomobject ifTrue:[
			tframe addChild: (self makeGeometry: node value).
			].
		].
	^ tframe.
		! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'sn 2/26/2018 12:30'!
makeHelperObject: tree

	| field fieldName name start end |
	tree linksDo:[ :node |  "Now linked list for performance"
		fieldName := node name.
		field := node value.
		fieldName = #TNodeName ifTrue:[
			start := field findDelimiters: '"' startingAt: 1.
			end := field findDelimiters: '"' startingAt: start+1.
			name := (field copyFrom: start+1 to: end-1).
			].
		fieldName = #TNodeTm ifTrue:[
			transform := self makeTransform: field.
			].
		].
	^ name.
		! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 15:02'!
makeMap: tree withOpacity: oNameAndMapOrNil

	| txtr fieldName field fname uOffset vOffset uTiling vTiling angle start end name extend dNameAndForm txtrform key |

" Create a texture based upon information found in the material field.
  oMapOrNil, if not nil, is the model's MAP_OPACITY for this model (as a  Form)."

	extend := nil.
	uOffset := 0.0.
	vOffset := 0.0.
	uTiling := 1.0.
	vTiling := 1.0.
	angle := 0.0.
	tree linksDo:[:node |  "linked list for performance"
		fieldName := node name.
		field := node value.
		fieldName = #TBitmap ifTrue:[ 
			start := field findDelimiters:'"' startingAt: 1.
			end := field findDelimiters: '"' startingAt: start+1.
			fname := field copyFrom: start+1 to: end-1.
			].
		fieldName = #TUvwUOffset ifTrue:[ uOffset := self makeNumber: field.].
		fieldName = #TUvwVOffset ifTrue:[ vOffset := self makeNumber: field.].
		fieldName = #TUvwUTiling ifTrue:[ uTiling := self makeNumber: field.].
		fieldName = #TUvwVTiling ifTrue:[ vTiling := self makeNumber: field.].
		fieldName = #TUvwAngle ifTrue:[ angle := self makeNumber: field.].
		fieldName = #TMapName ifTrue:[ name := field. 
				start := name findDelimiters:'$' startingAt: 1.
				start < (name size-1) ifTrue:[
					(name at:(start+1))=$c ifTrue:[extend := #colorKey.].
					(name at:(start+1))=$a ifTrue:[extend := #alpha].
					(name at:(start+1))=$f ifTrue:[extend := #fullBright].
					].
				].
		].
	fname ifNil:[^ nil].
	fname size = 0 ifTrue:[^ nil].
	"get file name from full path name - assumes file is a Windows file name"
	fname := fname reverse.
	end := fname findDelimiters:'\:'startingAt: 1.
	fname := (fname copyFrom: 1 to: end-1) reverse.

	fname ifNil:[^fname].

	oNameAndMapOrNil ifNil: [
		txtr := TTexture new initializeWithFileName: (filePath pathFromURI: fname) mipmap: true shrinkFit: true extension: extend.
	] ifNotNil: [
		dNameAndForm := self makeTextureForm: tree.
		"Check to see if we have this combination already"
		key := (TCryptoRandom hashMessage: dNameAndForm key, oNameAndMapOrNil key) asSymbol.
		txtr := Processor activeIsland at: key ifAbsent:[nil].
		txtr ifNotNil:[^txtr].
		txtrform := self applyOpacity: oNameAndMapOrNil value to: dNameAndForm value.
		txtr := TTexture new initializeWithForm: txtrform mipmap: true shrinkFit: true extension: #alpha.
		txtr registerGlobal: key.
	].

	txtr uvAngle: angle.
	txtr uvOffset: uOffset@vOffset.
	txtr uvScale: uTiling@vTiling.
	txtr objectName: name.
	^ txtr.
	


	! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 15:01'!
makeMaterial: tree

	| fieldName field material ambient diffuse specular shininess alpha txtr  subMaterialList twoSided mapDiffuseField oNameAndForm |

	alpha := 1.0.
	shininess := 50.0.
	material := TMaterial new.
	twoSided := false.
	txtr := nil.

	tree linksDo:[:node |
		fieldName := node name.
		field := node value.

		fieldName = #TMaterialName ifTrue:[material objectName: field.].
		fieldName = #TMaterialClass ifTrue:["unused"].
		fieldName = #TMaterialAmbient ifTrue:[ambient := self makeNumberArray: field.].
		fieldName = #TMaterialDiffuse ifTrue:[diffuse := self makeNumberArray: field].
		fieldName = #TMaterialSpecular ifTrue:[specular := self makeNumberArray: field].
		fieldName = #TMaterialShine ifTrue:[shininess := self makeNumber: field].
		fieldName = #TMaterialShinestrength ifTrue:["unused"].
		fieldName = #TMaterialTransparency ifTrue:[
						alpha := 1.0 - (self makeNumber: field)].
		fieldName = #TMaterialWiresize ifTrue:["unused"].
		fieldName = #TMaterialShading ifTrue:["unused"].
		fieldName = #TMaterialXpFalloff ifTrue:["unused"].
		fieldName = #TMaterialSoften ifTrue:["unused"].
		fieldName = #TMaterialXpType ifTrue:["unused"].
		fieldName = #TMapDiffuse ifTrue:[ mapDiffuseField := field ].
		fieldName = #TMapOpacity ifTrue: [ oNameAndForm := self makeTextureForm: field ].
		fieldName = #TNumsubmtls ifTrue:[ subMaterialList := OrderedCollection new.].
		fieldName = #TSubmaterial ifTrue:[ subMaterialList add: (self makeMaterial:field) ].
		fieldName = #TMaterialTwosided ifTrue:[ twoSided := true. ].
		].

	ambient ifNotNil:[
		ambient add: alpha.
		material ambientColor: ambient.
		].
	diffuse ifNotNil:[
		diffuse add: alpha.
		material diffuseColor: diffuse.
		].
	specular ifNotNil:[
		specular add: 1.0.
		material specularColor: specular.
		].
	material shininess: shininess.

	mapDiffuseField ifNotNil: [
		"Create the surface texture."
		txtr := self makeMap: mapDiffuseField withOpacity: oNameAndForm.
	].

	txtr ifNotNil:[

		material texture: txtr.

		txtr extension = #fullBright 
			ifTrue:[ material textureMode: GLReplace.
					material fullBright: true.
					Transcript show:'FULLBRIGHT';cr.
			]
			ifFalse:[ material textureMode: textureMode.].
		(txtr extension = #colorKey or:[txtr extension = #alpha])
			ifTrue:[material hasAlpha 
					ifFalse:[ 
						material diffuseColor:(material diffuseColor * 0.99)
						]
					].

		material uvAngle: txtr uvAngle.
		material uvOffset: txtr uvOffset.
		material uvScale: txtr uvScale.
		].

	material subMaterialList: subMaterialList.
	twoSided ifTrue:[ material cullFace: false].
	^ material.
! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 03:52'!
makeMaterials: tree

	| count cnter fieldName |

	count := 0.
	tree linksDo:[ :node |
		fieldName := node name.

		count=0 
			ifTrue: 
				[(fieldName = #TMaterialCount) ifTrue:[ count := (node value) asNumber].		
				count <= 0 ifTrue:[^self].
				materialList := OrderedCollection new.
				cnter := count.
				]
			ifFalse:[
				fieldName = #TMaterial ifTrue:[ 
					materialList add: (self makeMaterial: (node value)).].
					cnter := cnter -1.
					cnter<=0 ifTrue:[^self]
				].
		].! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 03:54'!
makeMesh: tree

	| fieldName field numVertices numFaces numTVertex numTVFaces tarray count matCount tv field2 field3 iterations |

	tvertices := nil.
	tfaces := nil.
	meshFaceNormals := nil.
	meshFaceVertexNormals := nil.
	
	tree linksDo:[ :node |
		fieldName := node name.
		field := node value.
		
		fieldName = #TMeshNumvertex ifTrue:[ 
			numVertices := self makeNumber: field.
			vertices := Vector3Array new: numVertices. 
			]
		ifFalse:[
		fieldName = #TMeshNumfaces ifTrue:[
			numFaces := self makeNumber: field.
			faces := IntegerArray ofSize: (numFaces*3).
			matIndex := IntegerArray ofSize: numFaces.
			]
		ifFalse:[
		fieldName = #TMeshVertexList ifTrue:[ 
			field ifNotNil:[
				count := 1.
				field do:[ :node2 |
					field2 := node2 value.
					tarray := self makeNumberArray: (field2).
					"NOTE: swap of 3DS y and z coordinates and negation of 3ds y coordinate"
					tarray := (Vector3 x:(tarray at: 2) y: (tarray at: 4) z:(tarray at: 3)negated)*scale.
					vertices at: count put: tarray.
					count := count+1.
					].
				].
			]
		ifFalse:[
		fieldName = #TMeshFaceList ifTrue:[
			count := 1.
			matCount := 1.
			iterations := 1.
			field do:[ :node2 |
				iterations=1 ifTrue: [field2 := node2 value].
				iterations=3 ifTrue: [field3 := node2 value.
					tarray := self makeNumberArray: (field2).
					faces at: count  put: (tarray at: 2).
					faces at: count+1 put: (tarray at: 3).
					faces at: count+2 put: (tarray at: 4).
					count := count+3.
					matIndex at: matCount put:(self makeNumber: field3).
					matCount := matCount + 1.
					iterations := 0.
					].
				iterations := iterations + 1.			
				]
			]
		ifFalse:[
		fieldName = #TMeshNumtvertex ifTrue:[
			numTVertex := self makeNumber:field.
			tvertices := Vector2Array new: numTVertex. 
			]
		ifFalse:[
		fieldName = #TMeshTvertlist ifTrue:[
			count := 1.
			field do:[ :node2 |
				field2 := node2 value.
				tarray := self makeNumberArray: (field2).
				tv := (Vector2 x:(tarray at:2) y:1-(tarray at:3)).
				tvertices at: count put: tv.
				count := count+1.
				].
			]
		ifFalse:[
		fieldName = #TMeshNumtvfaces ifTrue:[
			numTVFaces := self makeNumber: field.
			tfaces := IntegerArray ofSize:(numTVFaces*3). 
			]
		ifFalse:[
		fieldName = #TMeshNormals ifTrue:[
			"if we are recalculating anyway don't bother to get the normals 
				35% of time is spent getting numbers out of the fields" 
			(self recalculateFaceNormals: numFaces) 
				ifFalse: [meshFaceNormals := self makeFaceNormals: field.].
			(self recalculateFaceVertexNormals: numFaces) 
				ifFalse: [meshFaceVertexNormals := self makeFaceVertexNormals: field.].
			]
		ifFalse:[

		fieldName = #TMeshTfacelist ifTrue:[
			count := 1.
			field do:[ :node2 |
				field2 := node2 value.
				tarray := self makeNumberArray: (field2).
				tfaces at: count put: (tarray at: 2).
				tfaces at: count+1 put: (tarray at: 3).
				tfaces at: count+2 put: (tarray at: 4).
				count := count+3.
				].
			].
		]]]]]]]]
		].
! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 15:01'!
makeNodeName: tree

	| field fieldName name start end |
	tree linksDo:[ :node |
		fieldName := node name.
		fieldName = #TNodeName ifTrue:[
			field := node value.
			start := field findDelimiters: '"' startingAt: 1.
			end := field findDelimiters: '"' startingAt: start+1.
			name := (field copyFrom: start+1 to: end-1).
			].
		].
	^ name.
		! !

!TLoad3DSMax methodsFor: 'parse' stamp: ''!
makeNumber: string

	| start end substr|
	
	start := 1.
	end := 1.
	
	start := string findDelimiters: '1234567890-.' startingAt: end.
	end := string findDelimiters: (' ',Character tab asString) startingAt: start.
	substr := string copyFrom:start to: end-1.
	^ substr asNumber.
! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'jhk 4/29/2005 19:55'!
makeNumberArray: string

	| start farray parseStream |
	parseStream := ReadStream on: string.	
	farray := OrderedCollection new.
	start := 0.

	[start := string indexOfAnyOf: NumberSet startingAt: start+1 ifAbsent: [0].
		start = 0] whileFalse:[
			parseStream position: start-1.
			farray add: (Number readFrom: parseStream).
			start := parseStream position+1.
		].
	^ farray.! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'ar 7/6/2006 17:08'!
makeTextureForm: subtree

	| field fname start end stream node |

	"Create a Form from the BITMAP named in the given subtree."

	node := subtree detect: [:n| n name = #TBitmap ] ifNone: [ nil ].

	node ifNil: [^ nil].

	field := node value.
	start := field findDelimiters:'"' startingAt: 1.
	end := field findDelimiters: '"' startingAt: start+1.
	fname := field copyFrom: start+1 to: end-1.

	fname ifNil:[^ nil].
	fname size = 0 ifTrue:[^ nil].

	"get file name from full path name - assumes file is a Windows or Unix file name"
	fname := fname reverse.
	end := fname findDelimiters:'\:/'startingAt: 1.
	fname := (fname copyFrom: 1 to: end-1) reverse.

	fname ifNil:[^fname].

	[
		stream := filePath readOnlyFileNamed: (filePath pathFromURI: fname).
	     [ ^ fname -> (Form fromBinaryStream: stream) ] ensure: [ stream close. ].
	] on: Error do: [ :error|
		Transcript cr; show: 'ASE Loader cannot read texture ', fname, ': ', error description.
	].
	^nil

	


	! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 15:01'!
makeTransform: tree

	| field fieldName matrix v mOffset start end m90 |
	matrix := Matrix4x4 identity.
	tree linksDo:[ :node |
		fieldName := node name.
		field := node value.
		fieldName = #TNodeName ifTrue:[ "I put this here because groups don't give me a node name."
			start := field findDelimiters: '"' startingAt: 1.
			end := field findDelimiters: '"' startingAt: start+1.
			self objectName: (field copyFrom: start+1 to: end-1).
			].
		fieldName = #TTmRow0 ifTrue:[ 
			v := self makeNumberArray: field. 
			matrix a11: (v at: 1).
			matrix a21: (v at: 2).
			matrix a31: (v at: 3).
			].
		fieldName = #TTmRow1 ifTrue:[			
			v := self makeNumberArray: field. 
			matrix a12: (v at: 1).
			matrix a22: (v at: 2).
			matrix a32: (v at: 3).
			].
		fieldName = #TTmRow2 ifTrue:[
			v := self makeNumberArray: field. 
			matrix a13: (v at: 1).
			matrix a23: (v at: 2).
			matrix a33: (v at: 3).
			].
		fieldName = #TTmRow3 ifTrue:[ "********TRANSLATION FIELD******"
			v := self makeNumberArray: field. 
			matrix a14: (v at: 1) * scale.
			matrix a24: (v at: 2) * scale.
			matrix a34: (v at: 3) * scale .
			].
		fieldName = #TTmScale ifTrue:[ "*********SCALE FIELD************"
			v := self makeNumberArray: field.
			v at: 1 put:(1.0/v at:1).
			v at: 2 put:(1.0/v at:2).
			v at: 3 put:(1.0/v at:3).
			matrix a11: matrix a11*(v at:1).
			matrix a21: matrix a21*(v at:1).
			matrix a31: matrix a31*(v at:1).

			matrix a12: matrix a12*(v at:2).
			matrix a22: matrix a22*(v at:2).
			matrix a32: matrix a32*(v at:2).

			matrix a13: matrix a13*(v at:3).
			matrix a23: matrix a23*(v at:3).
			matrix a33: matrix a33*(v at:3).
			].
				" I am not sure if I need this next matrix or not."
		fieldName = #TNodeObjectoffset ifTrue:[ 
			mOffset := self makeTransform: field.
			].
		].

	m90 := Matrix4x4 identity rotationAroundX: -90.
	matrix := (m90 composeWith: matrix) composeWith: m90 orthoNormInverse.

"	^ B3DMatrix4x4 identity translation: (matrix translation)."
	^ matrix.! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'tak 6/22/2004 21:30'!
nextFieldOn: aStream 
	| name value |
	name := self
				convertTName: (self nextTokenTo: DelimSet2 on: aStream).
	value := self nextTokenTo: DelimSet1 on: aStream.
	^ Array with: name with: value withBlanksTrimmed! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'tak 6/22/2004 22:57'!
nextOn: aStream 
	"Answer pair of {name. value}. name is a Symbol and value is a String.  
	First character of name is $*, all rest string is stored into value.  
	If name is not found, name is set as nil.  
	Answer nil if end of file"
	[aheadChar
		ifNil: [^ Array with: #EOF with: nil].
	aheadChar = $*
		ifTrue: [^ self nextFieldOn: aStream].
	aheadChar = ${
		ifTrue: [self nextTokenTo: DelimSet1 on: aStream.
			^ Array with: #BEGIN with: nil].
	aheadChar = $}
		ifTrue: [self nextTokenTo: DelimSet1 on: aStream.
			^ Array with: #END with: nil].
	"Skip dirt"
	self nextTokenTo: DelimSet1 on: aStream] repeat! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'tak 6/22/2004 18:58'!
nextTokenTo: aCharactorSet on: aStream 
	"Fetch until one of the delimiter characters occured. The delimiter is  
	stored into aheadChar"
	| result |
	aStream atEnd
		ifTrue: [aheadChar := nil.
			^ ''].
	result := aStream upToAnyOf: aCharactorSet.
	aheadChar := result last.
	^ result allButLast! !

!TLoad3DSMax methodsFor: 'accessing' stamp: 'das 1/12/2006 11:01'!
objectName

	^ objectName.! !

!TLoad3DSMax methodsFor: 'accessing' stamp: 'das 1/12/2006 11:00'!
objectName: on

	objectName := on.! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'das 7/21/2006 20:25'!
oldCalcVertexNormals: faceNormals

"Calculate the normal for each face vertex."
	| vertexNormals vertexCount iCount faceIndex1 faceIndex2 edges index1 fv1 fv2 count countMax|

	"This is O(n-squared) in faces".
	faces size > 1000 ifTrue: [
		Transcript cr; show: 'Calculating vertex normals (', faces size, ' faces).'
	].

"Initialize the vertexNormals with the owning face normal."
	vertexNormals := Vector3Array  new: (faces size).
	vertexCount := IntegerArray new: (faces size).
	(self primCalcVertexNormals: vertexNormals vertexCount: vertexCount faces: faces faceNormals: faceNormals shadeAngleCos: shadeAngleCos)
		ifNotNil:[^vertexNormals].
	vertexNormals := Vector3Array  new: (faces size).
	vertexCount := IntegerArray new: (faces size).
	iCount := 1.
	1 to: faces size by: 3 do:[ :i |
		vertexNormals at: i put: (faceNormals at: iCount).
		vertexNormals at: i+1 put: (faceNormals at: iCount).
		vertexNormals at: i+2 put: (faceNormals at: iCount).
		iCount := iCount+1.].
"vertexCount is set to zero at the moment. It should be 1, but we know that it is always 1 greater than whatever our end result will be below."

	countMax := (faces size) // 20.
	count := 0.
"'Calculating Vertex Normals'
	displayProgressAt: Sensor cursorPoint
	from: 0 to: ((faces size) + 1)
	during: [:bar |"

	1 to: faces size do:[:i |
		i+1 to: faces size do:[:j |
			(faces at: i) = (faces at: j) ifTrue:[
				faceIndex1 := 1 + ((i-1)//3).
				faceIndex2 := 1 + ((j-1)//3).
				(self testNorm: (faceNormals at: faceIndex1) norm: (faceNormals at: faceIndex2))
					ifTrue:[
						vertexNormals at: i put: ((vertexNormals at: i) + (faceNormals at: faceIndex2)).
						vertexCount at: i put: 1+ (vertexCount at: i).
						vertexNormals at: j put: ((vertexNormals at: j) + (faceNormals at: faceIndex1)).
						vertexCount at: j put: 1+ ( vertexCount at: j).
						].
				].
			].
		count := count + 1.
		(count > countMax) ifTrue: ["bar value: i. " count := 0.].  
		].
"	bar value: ((faces size) + 1).
	]."
false ifTrue:[
"Build an edge dictionary"
	edges := Dictionary new.
	1 to: faces size by: 3 do:[ :i |
		edges at: (faces at:i)@(faces at:i+1) put: i@(i+1).
		edges at: (faces at: i+1)@(faces at: i+2) put: (i+1)@(i+2).
		edges at: (faces at: i+2)@(faces at: i) put: (i+2)@i].

"Now match edges"
	faceIndex2 := 1.
	1 to: faces size by: 3 do:[ :i |
		index1 := edges at: ((faces at:i+1)@(faces at:i)) ifAbsent:[ nil ].
		index1 ifNotNil:[ fv1 := i+1. fv2 := i.] ifNil:[
			index1 := edges at: ((faces at: i+2)@(faces at: i+1)) ifAbsent:[ nil. ].
			index1 ifNotNil:[ fv1 := i+2. fv2 := i+1.] ifNil:[
				index1 := edges at: ((faces at: i)@ (faces at: i+2)) ifAbsent:[ nil. ].
				index1 ifNotNil:[fv1 := i. fv2 := i+2.].
			].].

		index1 ifNotNil:[
			faceIndex1 := 1+((index1 x-1)//3).
			(self testNorm: (faceNormals at: faceIndex1) norm:(faceNormals at: faceIndex2))
				ifTrue:[
					vertexNormals at: fv1 put: ((vertexNormals at: fv1) + (faceNormals at: faceIndex1)).
					vertexNormals at: fv2 put: ((vertexNormals at: fv2) + (faceNormals at: faceIndex1)).
					vertexCount at: fv1 put: 1+(vertexCount at: fv1).
					vertexCount at: fv2 put: 1+(vertexCount at: fv2).
					fv1 := index1 x.
					fv2 := index1 y.
					vertexNormals at: fv1 put: ((vertexNormals at: fv1) + (faceNormals at: faceIndex2)).
					vertexNormals at: fv2 put: ((vertexNormals at: fv2) + (faceNormals at: faceIndex2)).].
					vertexCount at: fv1 put: 1+(vertexCount at: fv1).
					vertexCount at: fv2 put: 1+(vertexCount at: fv2).
				].						
		faceIndex2 := faceIndex2 + 1.
		].
].
" Calculate the average normal and normalize."
	1 to: vertexNormals size do:[ :i |
		(vertexNormals at: i) length = 0.0 ifFalse:[
		vertexNormals at:i put: ((vertexNormals at: i)/(1.0 + (vertexCount at: i)))normalized.].].

	^ vertexNormals.
! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'das 1/13/2006 10:10'!
parse: aStream 
	self reset.
	^self parseTree: aStream.
! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'jhk 5/12/2005 22:11'!
parse: aStream announcing: announcement
	| max tree |
	self reset.
	max := aStream size.
	announcement
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: max
		during: [:bar | tree := self parseTree: aStream bar: bar. bar value: max.].
	^tree.! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 02:33'!
parseFileNamed: fName 

	"MessageTally spyOn: [self new parseFileNamed: (FileDirectory  
	pathFrom: {FileDirectory default pathName. 'Content'. 'Plateau'.  
	'HPdemo.ASE'})]"

	| fileStream string tree |

	"Caution: This was using CrLfFileStream, which is very slow,
      but the plain stream appears to work correctly and is much faster.
      This may cause issues with ASE files produced on other platforms,
	 although it is fine for all current demo material (6/20/2006)... bgf."

	fileStream := (StandardFileStream readOnlyFileNamed: fName) ascii.
	
	[string := fileStream next: fileStream size]
		ensure: [fileStream close].

	tree := self parse: string readStream.
	^tree.! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'ar 4/4/2006 21:13'!
parseOld: string

	| tloc nloc tn name tree |
	tree := nil. "just to show this is accidental"
	tloc := string indexOfAnyOf: DelimSet1 startingAt: stringLoc ifAbsent:[string size+1].
	stringLoc := tloc+1.
" ---- Is this a label?  If not, then we are poorly formed. Bail as best you can."
	(string at: tloc) = $* ifTrue:[
		tree := OrderedCollection new.

" ---- Loop Here ---- "
		[stringLoc < string size] whileTrue:[

" ---- Find the field name or end of block or end of file."
			tloc := string indexOfAnyOf: DelimSet2 startingAt: stringLoc ifAbsent:[string size+1].
			(string at: tloc) = $} ifTrue:[ stringLoc := tloc + 1. ^tree. ].
			tloc >= string size ifTrue:[^tree.].
			(string at: tloc) = ${ ifTrue:[^tree.].
			tn := string copyFrom: stringLoc to: (tloc-1).
			name := 'T'. "Because this is Tea and because some field start with numbers."
			(tn findTokens: '_')do:[ :tkn | name := name, (tkn asLowercase capitalized)].
			name := name asSymbol.
			tree add: name.
" ---- Find the data."
			stringLoc := tloc+1.
			tloc := string indexOfAnyOf: DelimSet1 startingAt: stringLoc ifAbsent:[string size+1].

" ---- End of file."
			tloc > string size ifTrue:[ 
				tree add: (string copyFrom:stringLoc to: (tloc-1)). 
				stringLoc := tloc. 
				^tree.].

" ---- Found a new name or end of block- everything in between the last one and this must be useful."
			((string at: tloc) = $* or:[(string at: tloc)= $}])ifTrue:[ 
				nloc := string indexOf: Character cr startingAt: stringLoc ifAbsent:[string size+1].
				nloc > tloc ifTrue:[nloc := tloc].
				stringLoc <= (nloc-1) ifTrue:[
					tree add: (string copyFrom: stringLoc to: (nloc-1)).]
				ifFalse: [
					tree add:''].
				(string at: tloc)=$} ifTrue:[stringLoc := tloc. ^tree].
				stringLoc := tloc+1.
				].

" ---- Found a block - recurse."
			(string at: tloc) = ${ ifTrue: [
				stringLoc := tloc+1. 
				tree add: (self parseOld: string).
	" ---- Find the next name."
				stringLoc := string indexOfAnyOf: DelimSet3 startingAt: stringLoc+1 ifAbsent:[string size+1].
				stringLoc > string size ifTrue:[^tree].
				(string at: stringLoc) = $} ifTrue:[stringLoc := stringLoc+1. ^tree ].
				stringLoc := stringLoc+1.
				].
			].
		].
	^tree.

! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'tak 6/23/2004 09:59'!
parseOldFileNamed: fName 
	"MessageTally spyOn: [self new parseOldFileNamed: 
	(FileDirectory pathFrom: {FileDirectory default pathName. 'Content'. 
	'Plateau'. 'HPdemo.ASE'})]
	"
	| fileStream string |
	fileStream := (CrLfFileStream readOnlyFileNamed: fName) ascii.
	[string := fileStream next: fileStream size]
		ensure: [fileStream close].
	stringLoc := 1.
	^ self parseOld: string! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 03:44'!
parseTree: aStream
	| pair name value list count |
	list := LinkedList new.  "Uses linked list for performance - GC was thrashing on large OrderedCollections"
	count := 0.
	[
	pair := self nextOn: aStream.
	name := pair first.
	value := pair second.
	name = #EOF
		ifTrue: [^ list].
	name = #BEGIN
		ifTrue: ["Override last value"
				(list lastLink) value: (self parseTree: aStream)
				]
		ifFalse: [name = #END
				ifTrue: [^ list]
				ifFalse: [list addLast: (T3DSParseNode new name: name; value: value; yourself)]
			].
		count := count + 1.
		(count > 250) ifTrue: [count := 0.].
		] repeat.
! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'sn 2/26/2018 03:36'!
parseTree: aStream bar: bar 
	| pair name value list count |
	bar value: aStream position.
	list := LinkedList new.  "Uses linked list for performance - GC was thrashing on large OrderedCollections"
	count := 0.
	[pair := self nextOn: aStream.
	name := pair first.
	value := pair second.
	name = #EOF
		ifTrue: [^ list].
	name = #BEGIN
		ifTrue: ["Override last value"
				(list lastLink) value: (self parseTree: aStream bar: bar)
				]
		ifFalse: [name = #END
				ifTrue: [^ list]
				ifFalse: [list addLast: (T3DSParseNode name: name value: value)]
			].
		count := count + 1.
		(count > 250) ifTrue: [count := 0. bar value: aStream position.].
		] repeat.
		bar value: aStream position.! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'ar 9/30/2004 19:24'!
primCalcFaceNormals: argVertices faces: argFaces faceNormals: argFaceNormals

	<primitive: 'primitiveCalcFaceNormals' module: 'TLoad3DSMaxPlugin'>
	^ nil
	"TLoad3DSMaxPlugin doPrimitive: 'primCalcFaceNormals:faces:faceNormals:' withArguments: {argVertices. argFaces. argFaceNormals}."
! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'ar 9/30/2004 19:25'!
primCalcVertexNormals: argVertexNormals vertexCount: argVertexCount faces: argFaces faceNormals: argFaceNormals shadeAngleCos: shadeAngleCosValue


	<primitive: 'primitiveCalcVertexNormals' module: 'TLoad3DSMaxPlugin'>
	^ nil
	"TLoad3DSMaxPlugin doPrimitive: 'primCalcVertexNormals:vertexCount:faces:faceNormals:shadeAngleCos:' withArguments: {argVertexNormals. argVertexCount. argFaces. argFaceNormals. shadeAngleCosValue}."! !

!TLoad3DSMax methodsFor: 'accessing' stamp: 'bgf 6/29/2006 16:42'!
recalculateFaceNormals: fcs
	"Are face normals going to be recalculated rather than used from the file?"

	^ (fcs < 1000).    "Was 200000, but calculations are quite slow."! !

!TLoad3DSMax methodsFor: 'accessing' stamp: 'bgf 6/29/2006 16:41'!
recalculateFaceVertexNormals: fcs
	"are vertex normals going to be recalculated rather than being used from the file?"

	^ (fcs < 500).		"Was 20000, but calculation is very slow."! !

!TLoad3DSMax methodsFor: 'accessing' stamp: 'das 7/12/2006 21:47'!
reconstruct

"This method figures out how to reconstruct the mesh based solely upon the face information - material, vertex indices, vertex normals, and texture indices. It creates a group of sub-meshes which are then properly rendered. I realize that this could and should be broken down into many smaller methods, but I don't want to deal with creating throw-away classes that would only be used to help this method. On the other hand, I am very aware that there is a limit on the actual size of a method, so I might have to anyway."

	| faceNormals vertexFaceNormals vi ai ti mi vNorm fvi tvi avi mvi counter found faceGroup mm |

" First, make the face normals and vertex normals. "

	((self recalculateFaceNormals: (faces size)) or: [ meshFaceNormals = nil]) "not a really huge model or didn't get any normals in the ase file"
			 ifTrue:  [ faceNormals := self calcFaceNormals.]  "compute them"
			 ifFalse: [ Transcript show: 'Faces:', (faces size) asString, ' using file values for face normals'; cr. 
					  faceNormals := meshFaceNormals.
					  ].
	meshFaceNormals := nil. "Clear for next use."

	((self recalculateFaceVertexNormals: (faces size)) or: [ meshFaceVertexNormals = nil])  
			"if not a huge model or didn't get any normals in the ase file"
			ifTrue: [vertexFaceNormals := self calcVertexNormals: faceNormals.]
			ifFalse: [ Transcript show: 'Faces:', (faces size) asString, ' using file values for face vertex normals'; cr.
					vertexFaceNormals := meshFaceVertexNormals. 
					].
	meshFaceVertexNormals := nil.  "Clear for next use"

	tfaces ifNotNil:[self calcTextureVertices.].

"Faces now include the following information:
	- 1 Material index - matIndex
	- 3 Vertex indices - faces
	- 3 Vertex normals - vertexFaceNormals
	- 3 texture indices - tfaces
We build a new group of submeshes.
We first look at sub-groups according to materials. Then we compare vertices to build the final face tables.
A vertex is considered equal iff the index, normal, and texture index are identical. Otherwise, it must be aliased."

	vi := OrderedCollection ofSize: ((vertices size)).
	ai := OrderedCollection ofSize: ((vertices size)).
	ti := OrderedCollection ofSize: ((vertices size)).
	mi := OrderedCollection ofSize: ((vertices size)).
	vNorm := OrderedCollection ofSize: ((vertices size)).


1 to: faces size do:[ :i |
	found := false.
	fvi := 1+ (faces at: i).
	avi := fvi.
	tfaces ifNotNil:[tvi := 1 + (tfaces at: i).] 
		ifNil:[tvi := 1.].
	counter := 1+(i-1//3).
	mvi := matIndex at: counter.
	[found] whileFalse:[
		"Anything already in this slot?"
		(vi at: avi) ifNil:[
			vi at: avi put: fvi.
			ti at: avi put: tvi.
			vNorm at: avi put: (vertexFaceNormals at: i).
			mi at: avi put: mvi.
			found := true.
			].
		"Is what is there already the same as whet we have?"
		found ifFalse:[
			((vi at: avi) = fvi and:[
				(ti at: avi) = tvi and:[
					(mi at: avi) = mvi and:[
						self compare: (vNorm at: avi) 
							and: (vertexFaceNormals at: i) 
							within:0.0001.
							]
						]
					]) ifTrue:[
						faces at: i put: (avi-1).
						found := true.
						].
			].
		" There is something already there, but it is the wrong thing, we need an alias."
		found ifFalse:[
			(ai at: avi) 
				ifNil:[
					ai add: nil.
					vi add: fvi.
					ti add: tvi.
					vNorm add: (vertexFaceNormals at: i).
					mi add: mvi.
					ai at: avi put: vi size.
					faces at: i put: (vi size)-1.
					found := true.
					] 
				ifNotNil:[
					avi := ai at: avi.
					].
				].
			].
		].
"

The following commented out code fragment does the naive job of just adding all of the face information without looking for redundency. It is commented out because IT SHOULD NOT BE USED!! It is here just for reference, because someday, someone will get totally lost in what is going on here and this is a quick and dirty way to simplify the problem. The results are wrong of course, but at least it is a start.
-------
1 to: faces size do:[ :i |						
	vi add: 1+ (faces at:i).
	tfaces ifNotNil:[ ti add: 1 +( tfaces at: i). ] ifNil:[ ti add: 1.].
	vNorm add: (vertexFaceNormals at: i).
	faces at: i put: i-1.].
-------"


"Now clean up the arrays and convert them. The arrays I need to construct here are:
	vertexArray - the actual 3D vertex array including aliases.
	aliasArray - the alias index array - this will be used for vector based animations
	vertexNorms - the vertex normals
	textureArray - the u,v texture coordinate array
which are all the same length, and
	faceGroup - the collection of all faceGroups - check the associated material for alphas.
	"

	vi := self vertexIndexToArray: vi.
	ai := self aliasIndexToArray: ai.
	vNorm := self vertNormToArray: vNorm.
	tfaces ifNotNil:[ti := self textureIndexToArray: ti.] ifNil:[ ti := nil.].

	faceGroup := self facesToFaceGroup: faces materials: matIndex.
	mm := TMesh triMeshWithVertices: vi alias: ai norms: vNorm textureUV: ti faceGroups: faceGroup material: (materialRef ifNotNil:[materialList at: 1+materialRef]).    
	mm localTransform: self localTransform.
	mm objectName: self objectName.
	^ mm.! !

!TLoad3DSMax methodsFor: 'initialize' stamp: 'tak 6/22/2004 22:50'!
reset
	stringLoc := 1.
	aheadChar := $ . "anything ok"
! !

!TLoad3DSMax methodsFor: 'accessing' stamp: ''!
scale

	^ scale.! !

!TLoad3DSMax methodsFor: 'accessing' stamp: ''!
scale: scl

	scale := scl.! !

!TLoad3DSMax methodsFor: 'accessing' stamp: ''!
shadeAngle

	^ shadeAngle radiansToDegrees.! !

!TLoad3DSMax methodsFor: 'accessing' stamp: ''!
shadeAngle: angle

	shadeAngle := angle degreesToRadians.
	shadeAngleCos := shadeAngle cos.! !

!TLoad3DSMax methodsFor: 'parse' stamp: 'jhk 4/28/2005 18:20'!
showArray: array entries: depth title: title

		Transcript show: title; cr.
		1 to: depth do:[ :j | 
			Transcript show: (array at: j) asString; cr.
			]

! !

!TLoad3DSMax methodsFor: 'parse' stamp: ''!
showTree: tree

	self showTree: tree depth: 0.! !

!TLoad3DSMax methodsFor: 'parse' stamp: ''!
showTree: tree depth: depth

	| label |
	tree class = OrderedCollection ifTrue:[
	1 to: tree size by: 2 do:[:i |
		1 to: depth do:[ :j | Transcript show:'---|'.].
		label := tree at: i.
		Transcript show: (tree at: i) asString; cr.
		label = #TMeshVertexList ifFalse:[
		label = #TMeshFaceList ifFalse:[
		label = #TMeshTvertlist ifFalse:[
		label = #TMeshTfacelist ifFalse:[
		label = #TSubmaterial ifFalse:[
		self showTree: (tree at:(i+1)) depth: depth+1.].].].].].].].
! !

!TLoad3DSMax methodsFor: 'parse' stamp: ''!
testNorm: n1 norm: n2 

	" Test to see if the angle between the surfaces of n1 and n2 is greater than the shade angle, which is equivalent to being less than the cos of the shade angle. This means that we look at the negative of the dot product because the cos between the normals is the negative of the cos of the angle between the surfaces."
	^ (n1 dot: n2) negated < shadeAngleCos.! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'das 1/12/2006 10:19'!
textureIndexToArray: ti
	|textureArray|

	textureArray := Vector2Array ofSize: ti size.
	1 to: ti size do:[ :i |
		(ti at: i) ifNotNil:[
			textureArray at: i put: (tvertices at: (ti at:i)).]
		ifNil:[
			textureArray at: i put: Vector2 new.].
		].
	^ textureArray.! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'das 1/12/2006 10:19'!
textureTransform: tuv material: mat

	| uv |
	uv := tuv.
	mat texture ifNotNil:[
		"Magic mumbo jumbo goes here. Trial and error, and some intuition go a long way."
		uv := Point x: (uv x - 0.5 - mat uvOffset x) y: (uv y - 0.5 + mat uvOffset y).
		"uv := uv asPoint - 0.5 + mat uvOffset."
		mat uvAngle = 0.0 ifFalse:[
			uv := uv rotateBy: mat uvAngle about: 0@0.].
		uv := mat uvScale * uv + 0.5.
		].
	^ Vector2 x: uv x y: uv y.! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'das 1/12/2006 10:20'!
vertNormToArray: vi

	| vertexNormals |

	vertexNormals := Vector3Array ofSize: vi size.
	1 to: vi size do:[ :i |
		(vi at: i) ifNil:[
			vertexNormals at: i put: (Vector3 new).]
		ifNotNil:[
			vertexNormals at: i put: (vi at: i).].
		].
	^ vertexNormals.
			
		! !

!TLoad3DSMax methodsFor: 'convert' stamp: 'das 1/12/2006 10:20'!
vertexIndexToArray: vi

	| vertexArray |

	vertexArray := Vector3Array ofSize: (vi size).
	1 to: vi size do:[ :i |
		(vi at: i) ifNil:[
			vertexArray at: i put: Vector3 new.]
		ifNotNil:[
			vertexArray at: i put: (vertices at: (vi at: i)).
			].
		].
	^ vertexArray.! !

TFrame subclass: #TLoadMDL
	instanceVariableNames: 'textures'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Croquet-Import'!

!TLoadMDL commentStamp: '<historical>' prior: 0!
TLoadMDL is used to load Alice models.!

!TLoadMDL methodsFor: 'parsing' stamp: 'jcg 6/3/2006 23:37'!
buildMesh: vertices faces: faces normals: vtxNormals textureUV: vtxTexCoords textureMap: txtr
	| mat mesh |
	vertices ifNil:[^nil].
	faces ifNil:[^nil].
	vertices size < 3 ifTrue:[^nil].
	faces size < 1 ifTrue:[^nil].
	mat := TMaterial new.
	mat ambientColor: #(0.9 0.9 0.9 1).
	mat diffuseColor: #(0.9 0.9 0.9 1).
	txtr ifNotNil:[
		mat texture: txtr.
		mat textureMode: GLModulate.
	].

	vertices *= self scaleFactor.
	mesh := TMesh 
				triMeshWithVertices: vertices
				alias: nil 
				norms: vtxNormals 
				textureUV: vtxTexCoords
				faceGroups: {1. faces} 
				material: mat.
	^mesh! !

!TLoadMDL methodsFor: 'parsing' stamp: 'ar 1/13/2006 15:09'!
fixTransformsIn: anActor
	"Traverse anActor and all its children checking for non-orthonormal transforms.
	Move these down until we get to the leaves and there transform the meshes (if any)."
	| tfm offset rowX rowY rowZ scaleX scaleY scaleZ scale |
	tfm := anActor localTransform copy.
	"Check if the transform is *really* weird"
	(tfm a41 = 0.0 and:[tfm a42 = 0.0 and:[tfm a43 = 0.0 and:[tfm a44 = 1.0]]]) ifFalse:[
		self error: 'Cannot fix transform'.
	].

	"Remember offset of the transform"
	offset := tfm translation.
	tfm translation: 0@0@0.

	"Compute unit vectors to see if their length changes"
	rowX := tfm localPointToGlobal: 1@0@0.
	rowY := tfm localPointToGlobal: 0@1@0.
	rowZ := tfm localPointToGlobal: 0@0@1.

	"Normalize"
	scaleX := 1.0 / rowX length.	rowX *= scaleX.
	scaleY := 1.0 / rowY length.	rowY *= scaleY.
	scaleZ := 1.0 / rowZ length.	rowZ *= scaleZ.

	"Orthogonalize (later...)"
	(rowX dot: rowY) < 0.01 ifFalse:[self error: 'Not orthogonal'].
	(rowY dot: rowZ) < 0.01 ifFalse:[self error: 'Not orthogonal'].
	(rowZ dot: rowX) < 0.01 ifFalse:[self error: 'Not orthogonal'].

	"Create new orthonormal transform"
	tfm := Matrix4x4 identity.
	tfm a11: rowX x; a12: rowX y; a13: rowX z.
	tfm a21: rowY x; a22: rowY y; a23: rowY z.
	tfm a31: rowZ x; a32: rowZ y; a33: rowZ z.
	tfm translation: offset.
	anActor localTransform: tfm.

	"See if this a mesh"
	(anActor isKindOf: TMesh) ifTrue:[
		"Then rescale its vertices"
		scale := scaleX@scaleY@scaleZ negated.
		anActor vertices: (anActor vertices collect:[:vtx| vtx * scale]).
		anActor normals: (anActor normals collect:[:vtx| vtx x @ vtx y @ vtx z negated]).
		"And compute the bound spheres"
		anActor initBounds.
	].

	"Scale children (using a matrix since we'll orthonormalize them later) and recurse"
	scale := Matrix4x4 withScale: scaleX@scaleY@scaleZ.
	anActor frameChildren ifNotNil:[
		anActor frameChildren do:[:child|
			child translation: child translation x negated@ child translation y @ child translation z negated.
			child localTransform: (scale composeWith: child localTransform).
			self fixTransformsIn: child.
		].
	].
! !

!TLoadMDL methodsFor: 'as yet unclassified' stamp: 'pbm 8/15/2007 11:01'!
frame
	"for compatability"
	^ self! !

!TLoadMDL methodsFor: 'parsing' stamp: 'ar 10/8/2002 02:35'!
getChildNamed: childName from: parent

	| theChild nameParts |
	(parent objectName sameAs: childName) ifTrue: [ ^parent].

	nameParts := childName findTokens: ' '.
	(nameParts first sameAs: parent objectName) ifTrue: [nameParts removeFirst].
	theChild := parent.
	nameParts do: [:aName|
		theChild := theChild frameChildren detect:[:any| any objectName sameAs: aName].
	].
	^ theChild.! !

!TLoadMDL methodsFor: 'initialize' stamp: 'ar 1/15/2006 14:59'!
initialize
	textures := Dictionary new.! !

!TLoadMDL methodsFor: 'parsing' stamp: 'sn 2/26/2018 02:12'!
loadMeshFromFile: meshFile texture: textureName
	"Load this object's mesh from the specified file"
	| words texture |
	textureName ifNotNil:[
		texture := textures at: textureName 
			ifAbsentPut:[TTexture new initializeWithFileName: textureName].
	].
	meshFile ifNil:[^nil].
	words := (meshFile findTokens: '.').
	((words last) = 'vfb') 
		ifTrue: [^self parseVFBFile: meshFile texture: texture].

	((words last) = 'obj') 
		ifTrue: [^self parseOBJFile: meshFile texture: texture].
	^nil! !

!TLoadMDL methodsFor: 'parsing' stamp: 'sn 2/26/2018 02:15'!
parseFileNamed: filename
	"Creates a new actor using the specification from the given file"
	| aFile words line startSubstr index parent name texture meshFile matrix baseActor newActor fileVersion pos |

	words := (filename findTokens: '.').
	aFile :=  (MultiByteFileStream readOnlyFileNamed: filename) ascii;
	 wantsLineEndConversion: true;
	 yourself.

	"Check what version this mdl file is"
	line := aFile upTo: (Character cr).
	line := aFile upTo: (Character cr).
	line := aFile upTo: (Character cr).

	((line truncateTo: 7) = 'version')
		ifTrue: [ fileVersion := 1 ]
		ifFalse: [ fileVersion := 0 ].

	[ line := aFile upTo: (Character cr).
	aFile atEnd] whileFalse:[
		words := line findTokens: '='.
		"See if we're creating a new object"
		(((words size) > 1) and: [ ((words at: 2) beginsWith: ' _MakeObject')
			or: [ (words at: 2) beginsWith: ' Alice.MakeObject' ] ]) ifTrue: [
			(fileVersion = 0) ifTrue: [
				words := line findTokens: #,.
				parent := (words at: 2) withBlanksTrimmed.
				name := (((words at: 3) withBlanksTrimmed) findBetweenSubStrs: '"') at: 1.
			] ifFalse: [
				name := (words at: 1) truncateTo: (((words at: 1) size) - 1).
				parent := ((words at: 3) findTokens: #,) at: 1.
			].

			"Now pull in the texture to use"
			startSubstr := name , '.SetTexture'.
			pos := aFile position.
			texture := nil.
			[aFile atEnd or:[texture notNil]] whileFalse:[
				line := aFile upTo: (Character cr).
				(line beginsWith: startSubstr) ifTrue:[
					texture := (line findBetweenSubStrs: '"') at: 2.
					texture := self spell: texture in: aFile directory]].
			texture ifNil:[aFile position: pos].

			"Read the composite matrix to use"
			startSubstr := name , '._SetLocalTransformation'.
			pos := aFile position.
			matrix := nil.
			[aFile atEnd or:[matrix notNil]] whileFalse:[
				line := aFile upTo: (Character cr).
				(line beginsWith: startSubstr) ifTrue:[
					matrix := Matrix4x4 new.
					words := line findBetweenSubStrs: ',()'.
					words removeAllSuchThat: [:str | str = ' '].
					index := words size - 15.
					1 to: 4 do:[:j|
						1 to: 4 do:[:i|
							matrix at: i at: j put: 
								((words at: index) withBlanksTrimmed) asNumber.
							index := index + 1.]].
				].
			].
			matrix ifNil:[aFile position: pos].

			"Read the mesh file to use"
			startSubstr := 'LoadGeometry'.
			pos := aFile position.
			meshFile := nil.
			[aFile atEnd or:[meshFile notNil]] whileFalse:[
				line := aFile upTo: (Character cr).
				(line beginsWith: startSubstr) ifTrue:[
					meshFile := (line findBetweenSubStrs: '"') at: 2.
					meshFile := (aFile directory pathName), FileDirectory slash, meshFile.
				].
			].
			meshFile ifNil:[aFile position: pos].

			"Now build the actor name"
			words := name findTokens: '.'.
			name := words last.
			name at: 1 put: ((name at: 1) asLowercase).

			"Now build the parent name"
			parent := parent copyReplaceAll: '.' with: ' '.

			"Now create the object"
			(parent = 'None') ifTrue: [
				baseActor := self loadMeshFromFile: meshFile texture: texture.
				baseActor ifNil:[
					baseActor := TFrame new.
					baseActor visible: false].
				baseActor objectName: name.
				matrix ifNotNil:[baseActor localTransform: matrix].
				baseActor translation: baseActor translation * self scaleFactor.
				baseActor solid: false.
				"end base actor creation"
			] ifFalse: [
				newActor := self loadMeshFromFile: meshFile texture: texture.
				newActor ifNil:[
					newActor := TFrame new.
					newActor visible: false].
				newActor objectName: name.
				parent := self getChildNamed: parent from: baseActor.
				parent addChild: newActor.
				matrix ifNotNil:[newActor localTransform: matrix].
				newActor translation: newActor translation * self scaleFactor.
				newActor solid: false.
				"end new actor with parent"
			].
		"end MakeObject parsing"
		].

	]. "end file parsing"

	aFile close.
	self fixTransformsIn: baseActor.
	^baseActor! !

!TLoadMDL methodsFor: 'parsing' stamp: 'ar 1/13/2006 15:09'!
parseOBJFile: filename texture: txtr
	"Read in a mesh from the obj file"
	| aFile pos line words oldWords triple vCount fCount index vertices vtxTexCoords vtxNormals faces u v |
	aFile := (CrLfFileStream readOnlyFileNamed: filename) ascii.
	line := aFile upTo: (Character cr).
	words := line findTokens: ' '.
	vCount := 0.
	"Count the vertices"
	[ (words at: 1) = 'v' ] whileTrue: [
		vCount := vCount + 1.
		line := aFile upTo: (Character cr).
		words := line findTokens: ' '.
	].

	"Now reset the file and grab the actual data"
	aFile position: 0.

	"Create the vertex array"
	vertices := Vector3Array new: vCount.

	"Read in the vertices"
	line := aFile upTo: (Character cr).
	words := line findTokens: ' '.

	index := 1.
	[ (words at: 1) = 'v' ] whileTrue: [ 
		vertices at: index put:
				(Vector3 x: ((words at: 2) asNumber)
							y: ((words at: 3) asNumber)
							z: ((words at: 4) asNumber)).
		index := index + 1.
		line := aFile upTo: (Character cr).
		words := line findTokens: ' '.
	].

	"Read in the texture coordinates"
	index := 1.
	vtxTexCoords := Texture2Array new: vCount.
	[ (words at: 1) = 'vt' ] whileTrue: [
		u := (words at: 2) asNumber.
		v := (words at: 3) asNumber.
		v := 1.0 + (0.0 - v).
		vtxTexCoords at: index put: (Vector2 u: u v: v).
		index := index + 1.
		line := aFile upTo: (Character cr).
	 	words := line findTokens: ' '.
	].

	"Read in the normals"
	index := 1.
	vtxNormals := Vector3Array new: vCount.
	[ (words at: 1) = 'vn' ] whileTrue: [
		vtxNormals at: index
					put: (Vector3 x: ((words at: 2) asNumber)
									 y: ((words at: 3) asNumber)
									 z: ((words at: 4) asNumber)).
		index := index + 1.
		line := aFile upTo: (Character cr).
	 	words := line findTokens: ' '.
	].

	pos := aFile position.
	oldWords := words.

	"Count the faces"
	fCount := 0.
	[ (words size) > 0 ] whileTrue: [
		((words at: 1) = 'f') ifTrue: [fCount := fCount + 1 ].
		line := aFile upTo: (Character cr).
		words := line findTokens: ' '.
	].

	"Create the faces array"
	faces := WriteStream on: (IntegerArray new: fCount*3).

	aFile position: pos.
	words := oldWords.

	"Read in the faces"
	[ (words size) > 0 ] whileTrue: [
		((words at: 1) = 'f') ifTrue: [ 
			triple := (words at: 2) findTokens: '/'.
			faces nextPut: (triple at: 1) asNumber.

			triple := (words at: 3) findTokens: '/'.
			faces nextPut: (triple at: 1) asNumber.

			triple := (words at: 4) findTokens: '/'.
			faces nextPut: (triple at: 1) asNumber.
		].
		line := aFile upTo: (Character cr).
	 	words := line findTokens: ' '.
	].
	faces := faces contents.

	aFile close.
	^self buildMesh: vertices faces: faces normals: vtxNormals textureUV: vtxTexCoords textureMap: txtr.
! !

!TLoadMDL methodsFor: 'parsing' stamp: 'ar 1/13/2006 15:09'!
parseVFBFile: filename texture: txtr
	"Read in a mesh from the vfb file"
	| aFile w x y z bytes version vCount fCount verticesPerFace vertices vtxNormals vtxTexCoords faces |
		aFile := (StandardFileStream readOnlyFileNamed: filename) binary.
	"Read in the version number"
	bytes := aFile next: 4.
	version := bytes unsignedLongAt: 1 bigEndian: false.
	(version = 1) ifTrue: [
		"Read in the number of vertices"
		bytes := aFile next: 4.
		vCount := bytes unsignedLongAt: 1 bigEndian: false.
	] ifFalse: [
		vCount := version.
		version := 0.
	].

	(vCount <= 0) ifTrue: [
		aFile close.
		^nil
	].
	vertices := Vector3Array new: vCount.
	vtxNormals := Vector3Array new: vCount.
	vtxTexCoords := Texture2Array new: vCount.
	"Read in the vertices"
	1 to: vCount do: [:i | bytes := aFile next: 32.
		x := Float fromIEEE32Bit: (bytes unsignedLongAt: 1 bigEndian: false).
		y := Float fromIEEE32Bit: (bytes unsignedLongAt: 5 bigEndian: false).
		z := Float fromIEEE32Bit: (bytes unsignedLongAt: 9 bigEndian: false).
		vertices at: i put: (Vector3 x: (x negated) y: y z: z).
		x := Float fromIEEE32Bit: (bytes unsignedLongAt: 13 bigEndian: false).
		y := Float fromIEEE32Bit: (bytes unsignedLongAt: 17 bigEndian: false).
		z := Float fromIEEE32Bit: (bytes unsignedLongAt: 21 bigEndian: false).
		vtxNormals at: i put: (Vector3 x: (x negated) y: y z: z).
		x := Float fromIEEE32Bit: (bytes unsignedLongAt: 25 bigEndian: false).
		y := Float fromIEEE32Bit: (bytes unsignedLongAt: 29 bigEndian: false).
		y := 1.0 + (0.0 - y).
		vtxTexCoords at: i put: (Vector2 u: x v: y).
	].

	"Read in the number of faces"
	bytes := aFile next: 4.
	fCount := bytes unsignedLongAt: 1 bigEndian: false.
	(fCount <= 0) ifTrue: [
		aFile close.
		^nil].

	"Read past the faceDataCount value"
	aFile next: 4.
	(version = 0) 
		ifTrue: [ verticesPerFace := 0. ] 
		ifFalse: [bytes := aFile next: 4.
				verticesPerFace := bytes unsignedLongAt: 1 bigEndian: false].

	(verticesPerFace = 0) ifTrue: [
		faces := WriteStream on: (IntegerArray new: fCount*3).
		1 to: fCount do: [: i |
			bytes := aFile next: 4.
			w := bytes unsignedLongAt: 1 bigEndian: false.
			(w = 3) ifTrue: [
				bytes := aFile next: 4.
				x := (bytes unsignedLongAt: 1 bigEndian: false).
				bytes := aFile next: 4.
				y := (bytes unsignedLongAt: 1 bigEndian: false).
				bytes := aFile next: 4.
				z := (bytes unsignedLongAt: 1 bigEndian: false).
				faces nextPut: z; nextPut: y; nextPut: x.
			] ifFalse: [
				1 to: w do: [: j | aFile next: 4]
			].
		].
		faces := faces contents.
	].
	(verticesPerFace = 3) ifTrue: [
		faces := WriteStream on: (IntegerArray new: fCount*3).
		1 to: fCount do: [: i |
			bytes := aFile next: 4.
			x := (bytes unsignedLongAt: 1 bigEndian: false).
			bytes := aFile next: 4.
			y := (bytes unsignedLongAt: 1 bigEndian: false).
			bytes := aFile next: 4.
			z := (bytes unsignedLongAt: 1 bigEndian: false).
			faces nextPut: z; nextPut: y; nextPut: x.
		].
		faces := faces contents.
	].

	aFile close.
	^self buildMesh: vertices faces: faces normals: vtxNormals textureUV: vtxTexCoords textureMap: txtr.! !

!TLoadMDL methodsFor: 'parsing' stamp: 'ar 10/8/2002 09:55'!
scaleFactor
	^3! !

!TLoadMDL methodsFor: 'parsing' stamp: 'ar 10/17/2004 22:59'!
spell: misspelled in: aDirectory
	"Correct spelling for a potentially mis-spelled file in aDirectory"
	| corrected |
	corrected := aDirectory fileNames detect:[:any| any sameAs: misspelled].
	^aDirectory fullNameFor: corrected! !
TLoad3DSMax initialize!
